function [z,fval,info] = DWempc(c,G,g,H,h,lb,ub,kmax,...
    tol,LPsolver,v0,saveHist)
% DWempc: Implementation of the Dantzig-Wolfe Decomposition Algorithm
% described in the paper: "A Dantzig-Wolfe Decomposition Algorithm for
% Linear Economic Model Predictive Control of Dynamically Decoupled
% Subsystems".
% -------------------------------------------------------------------------
% INFO
% -------------------------------------------------------------------------
% Solves the Block-Angular Linear Program:
% min.  min_{z} sum(c_j^T z_j,  j = 1,2,...,m)
% s.t.          G_j z_j <= G_j, j = 1,2,...,m
%               lb_j <= z_j,    j = 1,2,...,m
%               z_j  <= ub_j,   j = 1,2,...,m
%               sum(H_j h_j, j = 1,2,...,m) = h
%
% -------------------------------------------------------------------------
% SYNTAX
% -------------------------------------------------------------------------
% z = DWempc(c,G,g,H,h)
% [z,fval,info] = DWempc(c,G,g,H,h,lb,ub,kmax,tol,LPsolver,v0,saveHist)
%
% -------------------------------------------------------------------------
% INPUTS
% -------------------------------------------------------------------------
% c{j}                : linear costs associated with unit j
% (G{j},g{j})         : constraints associated with unit j
% H{j} and h{1}       : linking constraints
% lb{j} (optional)    : lower bound counstraints associated with unit j
% ub{j} (optional)    : upper bound counstraints associated with unit j
% kmax (optional)     : maximum nr. of iterations
% tol  (optional)     : tolerance structure with tol.aGAP and tol.rGAP
% v0   (optional)     : initial point(s)        [v_0{1} v_0{2} ... v_0{m};
%                                                v_1{1} v_1{2} ... v_1{m}
%                                                  .     .        .
%                                                  .     .        .
%                                                  .     .        .
%                                                v_q{1} v_q{2} ... v_q{m}]
%                      where q is the number of initial points
% saveHist (optional) : 1 if iteration information should be saved, 0
%                       otherwise

% -------------------------------------------------------------------------
% OUTPUTS
% -------------------------------------------------------------------------
% z{j}           : solution vectors
% fval           : optimal objective value
% info           : info struct consisting of
%                    info.cputime ~ cputime (parallel) computed via tic/toc
%                    info.iter    ~ nr. of iterations performed
%                    info.converged ~ 1 if converged, 0 otherwise
%                    info.phi       ~ final pricing problem objectives
%                    info.v         ~ initial and generated columns
%                    info.fU        ~ obj. upper bound at each iteration
%                    info.fL        ~ obj. lower bound at each iteration
%                    info.hist      ~ if saveHist = 1, this field contains
%                                     the iterates and columns generated  
%                                     at every iteration
%   
% ========================================================================
% Leo Emil Sokoler (leso@dtu.dk), 01.01.2014
% DTU Informatics
% ========================================================================

%% ========================================================================
%  DEFAULT INPUTS
% =========================================================================
% bounds
if(nargin < 7)
    lb = cell(length(c),1);
    ub = cell(length(c),1);
end


% max nr. of iterations
if(nargin < 8 || isempty(kmax))
    kmax = 100;
end

% tolerance
if(nargin < 9 || isempty(tol))
    tol.aGAP = 1e-4;
    tol.rGAP = 1e-4;
end

% solver
if(nargin < 10 || isempty(LPsolver))
    LPsolver.solver = @linprog;
    LPsolver.opts = [];
end

% history
if nargin < 12 || isempty(saveHist)
    saveHist = 0;
end

%% ========================================================================
%  ERROR CHECKS
% =========================================================================
% tolerance and solver
eMis = isfield(tol,'aGAP') && isfield(tol,'rGAP') ...
    && isfield(LPsolver,'solver') && isfield(LPsolver,'opts');
if(eMis == 0)
    error('Missing inputs: check tolerance and solver specifications');
end

% LP dimensions
M = length(c);
eNum = (length(H) == M) && (length(h) == 1) && (length(G) == M) && ...
    (length(g) == M) && (length(ub) == M) && (length(c) == M);

if(eNum == 0)
    error('Data error: check that the LP dimensions are consistent');
end

eDim = zeros(M,1);
for j = 1:M
    eDim(j) = (size(G{j},1) == size(g{j},1)) && ...
        (size(G{j},2) == size(c{j},1)) && ...
        (size(H{j},1) == size(h{1},1)) && (size(H{j},2) == size(c{j},1));
end
if(any(eDim == 0))
    error('Data error: check that the LP dimensions are consistent');
end

%% ========================================================================
%  INITIALIZE ALGORITHM
% =========================================================================
u_init = cell(1,M);
for j = 1:M
    u_init{1,j} = LPsolver.solver(-c{j},G{j},g{j},...
        [],[],lb{j},ub{j},[],LPsolver.opts);
    u_init{2,j} = LPsolver.solver(zeros(size(c{j})),G{j},g{j},...
        [],[],lb{j},ub{j},[],LPsolver.opts);
    if(isempty(u_init{1,j}))
        error(strcat('Infeasible: Subproblem ',num2str(j)));
    end
end

% user supplied
if nargin < 11 || isempty(v0)
    v_init = u_init;
else
    % check initial guess
    e_init = (size(v0,2) == M);
    
    if(e_init == 0)
        error('Data error: check the initial point(s)');
    end
    
    e_init = zeros(M,size(v0,1));
    for j = 1:M
        for k = 1:size(v0,1)
            e_init(j,k) = length(v0{k,j}) == length(c{j});
        end
    end
    
    if(any(e_init(:) == 0))
        error('Data error: check the initial point(s)');
    end
    
    % add initial guess
    v_init = [v0; u_init];
end

P = size(v_init,1);

%% ========================================================================
% PREALLOCATE
% =========================================================================
z = cell(M,1);
v = cell(kmax,M);

phi = zeros(M,kmax);
H_ = cell(kmax,M);
c_ = cell(kmax,M);

fL = zeros(kmax,1);
fU = zeros(kmax,1);
dA = zeros(kmax,1);
dR = zeros(kmax,1);
phiU = zeros(kmax,1);

if(saveHist)
    info.cell.z = cell(kmax,M);
end

%% ========================================================================
% INITIAL VERTICES
% =========================================================================
for i = 1:P
    for j = 1:M
        v(i,j) = v_init(i,j);
    end
end

for j = 1:M
    for i = 1:P-1
        H_{i,j} = H{j}*v{i,j};
        c_{i,j} = c{j}'*v{i,j};
    end
end

%% ========================================================================
% ITERATION PROCEDURE
% =========================================================================
converged = 0; k = 0; fL(1) = -inf; cputime = 0;
while(k < kmax && ~converged)
    k = k + 1;
    
     t0 = tic;
    % UPDATE MASTER PROBLEM
    for j = 1:M
        H_{P,j} = H{j}*v{P,j};
        c_{P,j} = c{j}'*v{P,j};
    end
    cputime = cputime + toc(t0)/M;
    
    t0 = tic;
    % SOLVE MASTER PROBLEM
    [lambda,alpha,beta,fval] = ...
        DWempcMaster(c_(1:P,:),H_(1:P,:),h{1},LPsolver);
    cputime = cputime + toc(t0);
    fU(k) = fval;
    
    % GET SOLUTION
    t0 = tic;
    l = 1;
    for j = 1:M
        z{j} = zeros(length(v{1,j}),1);
        for i = 1:P
            z{j} = z{j} + lambda(l)*v{i,j};
            l = l + 1;
        end
    end
    cputime = cputime + toc(t0)/M;
    
    if(saveHist)
        info.hist.z(k,:) = z;
    end
    
    % SOLVE SUBPROBLEM
    [v(P+1,:),phi(:,P),xi,tsub] = ...
        DWempcSubproblem(c,H,G,g,lb,ub,alpha,beta,v(P,:),LPsolver);
     cputime = cputime + max(tsub);
    
    % OBJECTIVE LOWER BOUND
    t0 = tic;
    fLs = -alpha'*h{1};
    cputime = cputime + toc(t0);
    
    t0 = tic;
    for j = 1:M
        u1 = -[xi{j}.ineqlin; xi{j}.lower; xi{j}.upper];
        u2 = [g{j}; lb{j}; ub{j}];
        fLs = fLs + u1'*u2;
    end
    cputime = cputime + toc(t0);
    
    if(k == 1)
        fL(k) = fLs;
    elseif(fLs > fL(k-1))
        fL(k) = fLs;
    else
        fL(k) = fL(k-1);
    end
    dA(k) = fU(k) - fL(k);
    dR(k) = dA(k)/(1+abs(fL(k)));
    
    % CHECK IF CONVERGED
    if((dR(k) <= tol.rGAP) && (dA(k) <= tol.aGAP))
        converged = true;
    end
    phiU(k) = max(abs(phi(:,P)));
    
    P = P + 1;
end

if(saveHist)
    info.hist.v = v(1:P,:);
end

info.fU = fU(1:k);
info.fL = fL(1:k);
phi = phi(:,1:P-1);
info.cputime = cputime;
info.phi = phi;
info.converged = converged;
info.iter = k;


function [lambda,alpha,beta,fval,exitflag,output] = ...
 DWempcMaster(c,H,h,LPsolver)

[P,M] = size(H);
m = length(H{1});
H_agg = zeros(m,P*M);
c_agg = zeros(P*M,1);
tmp = cell(P);
for i = 1:M
    tmp{i} = sparse(ones(1,P));
end
V_agg = blkdiag(tmp{:});
V_rhs = ones(M,1);
lb = zeros(M*P,1);
k = 1;
for j = 1:M
    for i = 1:P
        H_agg(:,k) = H{i,j};
        c_agg(k) = c{i,j};
        k = k + 1;
    end
end
[lambda,fval,exitflag,output,dual] = LPsolver.solver(c_agg,H_agg,h, ...
    V_agg,V_rhs,lb,[],[],LPsolver.opts);
if(isempty(lambda))
    error('Infeasible: Master problem');
end
alpha = abs(dual.ineqlin);
beta = abs(dual.eqlin);

function [v,phi,xi,t] = ...
    DWempcSubproblem(c,H,G,g,lb,ub,alpha,beta,v0,LPsolver)
M = length(c);
v = cell(M,1);
t = zeros(M,1);
phi = zeros(M,1);
xi = cell(M,1);
for j = 1:M
    c_sub = (c{j} + H{j}'*alpha);
    G_sub = G{j};
    g_sub = g{j};
    s_id = tic;
    [v{j},fval,~,~,xi{j}] = ...
        LPsolver.solver(c_sub,G_sub,g_sub,...
        [],[],lb{j},ub{j},v0{j},LPsolver.opts);
    t(j) = toc(s_id);
    phi(j) = fval + beta(j);
end