%% ========================================================================
% Test of DWempc: Energy system case study
% Written by: Leo Emil Sokoler, leso@dtu.dk
% 01/01/2014, DTU Informatics
% =========================================================================
rng(10);     % set fixed random state (change for different examples)
plotSol = 1; % display solution; disable for large problems

%% ========================================================================
% SYSTEM PARAMTERS
% =========================================================================
M = 8; % e.g. M = 2000 for a large-scale problem.
a = 20; b = 80;
Tau = a + (b-a).*rand(M,1);

%% ========================================================================
% INTERNAL DW SOLVER SETTINGS
% =========================================================================
% solver needs "linprog" interface. Tested with cplex, gurobi and mosek 
% these solvers can be downloaded for free academic use
%
LPsolver.solver = @linprog;
LPsolver.opts = optimset('Display','off','Diagnostics','off',...
   'MaxIter',1e8);

% CPLEX suggested settings
% LPsolver.solver = @cplexlp;
% LPsolver.opts = cplexoptimset('Display','off','Diagnostics','off',...
%     'MaxIter',1e8,'Simplex','on');

tol.aGAP = 1e-2;
tol.rGAP = 1e-3;
kmax = 20;
if(plotSol)
    saveHist = 1;
else
    saveHist = 0;
end


%% ========================================================================
% SYSTEM SETUP
% =========================================================================
Ts = 5;
[A,B,C,Upsilon] = ESsystem(Tau,Ts);
nx = cell(M,1);
nu = cell(M,1);
ny = cell(M,1);
x0 = cell(M,1);

for j = 1:M
    [nx{j},nu{j}] = size(B{j});
    ny{j} = size(C{j},1);
    x0{j} = zeros(nx{j},1);
end

%% ========================================================================
% CONTROLLER SETUP
% =========================================================================
N = 60;
p = cell(M,1);
r = cell(M,1);
w = cell(M,1);
q = cell(1,1);
u_min = cell(M,1);
u_max = cell(M,1);
Du_min = cell(M,1);
Du_max = cell(M,1);
y_min = cell(M,1);
y_max = cell(M,1);
yT_min = cell(1,1);
yT_max = cell(1,1);
gamma_max = cell(M,1);
rho_max = cell(1,1);
eta_max = cell(M,1);
u_m1 = cell(M,1);
for j = 1:M
    u_m1{j} = zeros(nu{j},1);
    w{j} = 0.01*ones(nu{j},N);       % regularization
    p{j} = 1/Tau(j)*rand(nu{j},N);   % input price
    r{j} = 0*ones(ny{j},N);          % local soft violation price
    u_min{j} = zeros(nu{j},N);
    u_max{j} = 20*rand*ones(nu{j},N)/M;
    Du_min{j} = -u_max{j}/4;
    Du_max{j} =  u_max{j}/4;
    y_min{j} = -inf*ones(ny{j},N);
    y_max{j} = inf*ones(ny{j},N);
    gamma_max{j} = 0*ones(ny{j},N); % max local violation
    eta_max{j} = max(max(max(Du_max{j},-Du_min{j})))...
        *ones(nu{j},N);              % max input-rate movement
end
q{1} = 10*ones(1,N);                 % global soft violation price
yT_min{1} = zeros(1,N);
yT_max{1} = 5*ones(1,N);
rho_max{1} = 10*ones(1,N);           % max global violation
yT_min{1}(1:ceil(N/2)) = 0;
yT_min{1}(ceil(N/4):ceil(N/2)) = linspace(0,4,ceil(N/4)+1);
yT_min{1}(ceil(N/2):end) = 4;
yT_max{1} = yT_min{1}+1;

%% ========================================================================
% DW BLOCK-CONSTRAINED FORMULATION
% =========================================================================
[c,G,g,H,h,lb,ub] = DWempcSetupLP(A,B,C,Upsilon,x0,p,w,r,q,u_min,...
    u_max,Du_min,Du_max,y_min,y_max,yT_min,yT_max,gamma_max,...
    rho_max,eta_max,u_m1);

%% ========================================================================
% SOLVE USING CENTRALIZED SOLVER
% =========================================================================
c_ = cell2mat(c);
A_ = [blkdiag(G{:}); [H{:}]];
b_ = [cell2mat(g); cell2mat(h)];
lb_ = cell2mat(lb);
ub_ = cell2mat(ub);
tic;
[Z_,fval_] = LPsolver.solver(c_,A_,b_,[],[],lb_,ub_);
t_cen = toc;
pause(0.1);
z_ = cell(M,1);
k = 1;
for j = 1:M+1
    m = size(c{j},1);
    z_{j} = Z_(k:k+m-1);
    k = k + m;
end
[u_,x_,y_,yT_] = DWempcGetSol(z_,A,B,C,Upsilon,x0,N);

%% ========================================================================
% PLOT SOLUTION - CENTRALIZED SOLVER
% =========================================================================
if(plotSol)
    h_fig = figure(1);
    set(h_fig,'name','Solution: Centralized Algorithm');
    T = 0:Ts:N*Ts;
    subplot(2,1,1);
    str = cell(1,M);
    for i = 1:M
        plot(T(1:end-1),u_{i},'color',rand(1,3)); hold on; grid on;
        str{i} = strcat('unit ',num2str(i));
    end
    legend(str);
    ylabel('Input');
    subplot(2,1,2);
    plot(T,yT_{1}); grid on; hold on;
    plot(T(2:end),yT_min{1},'r--');
    plot(T(2:end),yT_max{1},'r--');
    legend('y_T','min/max','location','se');
    xlabel('Time [sec]');
    ylabel('Output');
end

%% ========================================================================
% SOLVE USING DANTZIG WOLFE DECOMPOSITON
% =========================================================================
[z,fval,info] = DWempc(c,G,g,H,h,lb,ub,kmax,tol,LPsolver,[],saveHist);
t_dw = info.cputime;

%% ========================================================================
% SOLVE USING DANTZIG WOLFE DECOMPOSITON
% =========================================================================
fprintf('\n========================================================== \n');
fprintf('Constraint matrix dimensions: (%g,%g) \n', size(A_,1), size(A_,2));
fprintf('Centralized Algorithm CPU-time: %g \n', t_cen);
fprintf('Dantzig-Wolfe Algorithm CPU-time (effective): %g \n', t_dw);
fprintf('Dantzig-Wolfe Algorithm #iterations: %g \n', info.iter);
fprintf('Dantzig-Wolfe Suboptimality: %g \n', (fval-fval_)/fval);
fprintf('========================================================== \n ');

%% ========================================================================
% PLOT SOLUTION (ITERATIVELY) - DW SOLVER
% =========================================================================
if(plotSol)
    h_fig = figure(2);
    set(h_fig,'name','Solution: Dantzig-Wolfe Decomposition Algorithm');
    T = 0:Ts:N*Ts;
    for k = 1:info.iter
        [u,x,y,yT] = DWempcGetSol(info.hist.z(k,:),...
            A,B,C,Upsilon,x0,N);
        subplot(2,1,1); cla;
        str = cell(1,M);
        rng(1);
        for i = 1:M
            plot(T(1:end-1),u{i},'color',rand(1,3)); hold on; grid on;
            str{i} = strcat('unit ',num2str(i));
        end
        legend(str);
        ylabel('Input');
        subplot(2,1,2); cla;
        plot(T,yT{1}); grid on; hold on;
        plot(T(2:end),yT_min{1},'r--');
        plot(T(2:end),yT_max{1},'r--');
        legend('y_T','min/max','location','se');
        xlabel('Time [sec]');
        ylabel('Output');
        % pause after each iteration and plot new solution
        fprintf('\nDW Solution at iteration %g/%g..... \n',k,info.iter);
        fprintf('Press any key to continue \n'); 
        pause
    end
end

%% PLOT DW ITERATION SEQUENCE BOUNDS
figure(3);
plot([1 info.iter],log([fval_ fval_])); grid on; hold on;
plot(1:info.iter,abs(log(info.fU)).*sign(info.fU),'r--'); 
plot(1:info.iter,abs(log(info.fL)).*sign(info.fL),'r--');
xlabel('#iteration'); axis tight;
ylabel('optimality gap (log transformed)');
legend('optimal value','upper/lower bound');

